"""
Functions for generating an IocInstance derived class from a
support module definition YAML file
"""

from __future__ import annotations

import builtins
from enum import Enum
from typing import Annotated, Any, Dict, Literal, Sequence, Tuple, Type, Union

from pydantic import (
    Field,
    create_model,
    field_validator,
    model_validator,
)
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined

from .globals import BaseSettings
from .support import Definition, EnumArg, IdArg, ObjectArg, Support
from .utils import UTILS

id_to_entity: Dict[str, Entity] = {}


class EnumVal(Enum):
    """
    An enum that is printed as its name only
    """

    def __str__(self):
        return self.name


class Entity(BaseSettings):
    """
    A baseclass for all generated Entity classes.
    """

    type: str = Field(description="The type of this entity")
    entity_enabled: bool = Field(
        description="enable or disable this entity instance", default=True
    )
    __definition__: Definition

    @model_validator(mode="after")  # type: ignore
    def add_ibek_attributes(cls, entity: Entity):
        """
        Whole Entity model validation
        """

        # find the id field in this Entity if it has one
        ids = {a.name for a in entity.__definition__.args if isinstance(a, IdArg)}

        entity_dict = entity.model_dump()
        for arg, value in entity_dict.items():
            if arg in ids:
                # add this entity to the global id index
                if value in id_to_entity:
                    raise ValueError(f"Duplicate id {value} in {list(id_to_entity)}")
                id_to_entity[value] = entity
            elif isinstance(value, str):
                # Jinja expansion of any of the Entity's string args/values
                setattr(entity, arg, UTILS.render(entity_dict, value))
        return entity

    def __str__(self):
        # if this entity has an id then its string representation is the value of id
        id_name = self.__definition__._get_id_arg()
        return getattr(self, id_name) if id_name else super().__str__()


def make_entity_model(definition: Definition, support: Support) -> Type[Entity]:
    """
    Create an Entity Model from a Definition instance and a Support instance.
    """

    def add_arg(name, typ, description, default):
        if default is None:
            default = PydanticUndefined
        args[name] = (
            typ,
            FieldInfo(description=description, default=default),
        )

    args: Dict[str, Tuple[type, Any]] = {}
    validators: Dict[str, Any] = {}

    # fully qualified name of the Entity class including support module
    full_name = f"{support.module}.{definition.name}"

    # add in each of the arguments as a Field in the Entity
    for arg in definition.args:
        full_arg_name = f"{full_name}.{arg.name}"
        arg_type: Any

        if isinstance(arg, ObjectArg):

            @field_validator(arg.name, mode="after")
            def lookup_instance(cls, id):
                try:
                    return id_to_entity[id]
                except KeyError:
                    raise ValueError(f"object {id} not found in {list(id_to_entity)}")

            validators[full_arg_name] = lookup_instance
            arg_type = object

        elif isinstance(arg, IdArg):
            arg_type = str

        elif isinstance(arg, EnumArg):
            # Pydantic uses the values of the Enum as the options in the schema.
            # Here we arrange for the keys to be in the schema (what a user supplies)
            # but the values to be what is rendered when jinja refers to the enum
            enum_swapped = {}
            for k, v in arg.values.items():
                enum_swapped[str(v) if v else str(k)] = k
            val_enum = EnumVal(arg.name, enum_swapped)  # type: ignore
            arg_type = val_enum

        else:
            # arg.type is str, int, float, etc.
            arg_type = getattr(builtins, arg.type)
        default = getattr(arg, "default", None)
        add_arg(arg.name, arg_type, arg.description, default)

    # add in the calculated values Jinja Templates as Fields in the Entity
    for value in definition.values:
        add_arg(value.name, str, value.description, value.value)

    # add the type literal which discriminates between the different Entity classes
    typ = Literal[full_name]  # type: ignore
    add_arg("type", typ, "The type of this entity", full_name)

    entity_cls = create_model(
        full_name.replace(".", "_"),
        **args,
        __validators__=validators,
        __base__=Entity,
    )  # type: ignore

    # add a link back to the Definition Instance that generated this Entity Class
    entity_cls.__definition__ = definition

    return entity_cls


def make_entity_models(support: Support):
    """
    Create Entity subclasses for all Definition instances in the given
    Support instance. Returns a list of the Entity subclasses Models.
    """

    entity_models = []
    entity_names = []

    for definition in support.defs:
        entity_models.append(make_entity_model(definition, support))
        if definition.name in entity_names:
            # not tested because schema validation will always catch this first
            raise ValueError(f"Duplicate entity name {definition.name}")
        entity_names.append(definition.name)

    return entity_models


def make_ioc_model(entity_models: Sequence[Type[Entity]]) -> Type[IOC]:
    """
    Create an IOC derived model, by setting its entities attribute to
    be of type 'list of Entity derived classes'.
    """

    entity_union = Union[tuple(entity_models)]  # type: ignore
    discriminated = Annotated[entity_union, Field(discriminator="type")]  # type: ignore

    class NewIOC(IOC):
        entities: Sequence[discriminated] = Field(
            description="List of entities this IOC instantiates"
        )

    return NewIOC


def clear_entity_model_ids():
    """Resets the global id_to_entity dict."""

    id_to_entity.clear()


class IOC(BaseSettings):
    """
    Used to load an IOC instance entities yaml file into a Pydantic Model.
    """

    ioc_name: str = Field(description="Name of IOC instance")
    description: str = Field(description="Description of what the IOC does")
    entities: Sequence[Entity]



import os
from pathlib import Path

from pytest import fixture
from ruamel.yaml import YAML
from typer.testing import CliRunner

# This must be above the following imports so that it takes effect before
# `globals.EPICS_ROOT` is imported (or anything built on top of it)
os.environ["EPICS_ROOT"] = str(Path(__file__).parent / "samples" / "epics")

# The `noqa`s on these imports are necessary because of the above
from ibek.__main__ import cli  # noqa: E402
from ibek.ioc import clear_entity_model_ids, make_entity_models  # noqa: E402
from ibek.support import Support  # noqa: E402

runner = CliRunner()


def run_cli(*args):
    result = runner.invoke(cli, [str(x) for x in args])
    if result.exception:
        raise result.exception
    assert result.exit_code == 0, result


def get_support(yaml_file: str) -> Support:
    """
    Get a support object from the sample YAML directory
    """
    # load from file
    d = YAML(typ="safe").load(yaml_file)
    # create a support object from that dict
    support = Support(**d)
    return support


@fixture
def templates():
    return Path(__file__).parent.parent / "src" / "ibek" / "templates"


@fixture
def samples():
    return Path(__file__).parent / "samples"


@fixture
def support_defs(samples):
    return samples / "support"


@fixture
def ioc_defs(samples):
    return samples / "iocs"


@fixture
def asyn_classes(support_defs):
    # clear the entity classes to make sure there's nothing left
    clear_entity_model_ids()

    asyn_support = get_support(support_defs / "asyn.ibek.support.yaml")

    # make entity subclasses for everything defined in it
    namespace = make_entity_models(asyn_support)

    # return the namespace so that callers have access to all of the
    # generated dataclasses
    return namespace


@fixture
def motor_classes(support_defs):
    # clear the entity classes to make sure there's nothing left
    clear_entity_model_ids()

    asyn_support = get_support(support_defs / "asyn.ibek.support.yaml")
    motor_support = get_support(support_defs / "motorSim.ibek.support.yaml")

    # make entity subclasses for everything defined in it
    namespace = make_entity_models(asyn_support)
    namespace.extend(make_entity_models(motor_support))

    # return the namespace so that callers have access to all of the
    # generated dataclasses
    return namespace



"""
Functions for rendering lines in the boot script using Jinja2
"""

from typing import Callable, List, Optional, Union

from .ioc import IOC, Entity
from .support import Comment, Script, Text, When
from .utils import UTILS


class Render:
    """
    A class for generating snippets of startup script / EPICS DB
    by using Jinja to combine snippet templates from support module
    definition yaml with substitution values supplied in ioc entity yaml
    """

    def __init__(self: "Render"):
        self.once_done: List[str] = []

    def render_text(
        self, instance: Entity, text: str, when=When.every, suffix=""
    ) -> str:
        """
        Add in the next line of text, honouring the once flag which will
        only add the line once per IOC.

        Jinja rendering of values/args has already been done in Entity.__post_init__
        but we pass all strings though jinja again to render any other jinja
        in the IOC (e.g. database and function entries)

        once uses the name of the definition + suffix to track which lines
        have been rendered already. The suffix can be used where a given
        Entity has more than one element to render once (e.g. functions)
        """

        if when == When.first.value:
            name = instance.__definition__.name + suffix
            if name not in self.once_done:
                self.once_done.append(name)
            else:
                return ""
        elif when == When.last.value:
            raise NotImplementedError("When.last not yet implemented")

        # Render Jinja entries in the text
        result = UTILS.render(instance, text)  # type: ignore

        if result == "":
            return ""

        return result + "\n"

    def render_script(self, instance: Entity, script_items: Script) -> Optional[str]:
        script = ""

        for item in script_items:
            if isinstance(item, Comment):
                comments = "\n".join(["# " + line for line in item.value.split("\n")])
                script += self.render_text(
                    instance, comments, item.when, suffix="comment"
                )
            elif isinstance(item, Text):
                script += self.render_text(
                    instance, item.value, item.when, suffix="text"
                )

        return script

    def render_pre_ioc_init(self, instance: Entity) -> Optional[str]:
        """
        render the startup script by combining the jinja template from
        an entity with the arguments from an Entity
        """
        pre_init = instance.__definition__.pre_init
        return self.render_script(instance, pre_init)

    def render_post_ioc_init(self, instance: Entity) -> Optional[str]:
        """
        render the post-iocInit entries by combining the jinja template
        from an entity with the arguments from an Entity
        """
        post_init = instance.__definition__.post_init
        return self.render_script(instance, post_init)

    def render_environment_variables(self, instance: Entity) -> Optional[str]:
        """
        render the environment variable elements by combining the jinja template
        from an entity with the arguments from an Entity
        """
        variables = getattr(instance.__definition__, "env_vars")
        if not variables:
            return None

        env_var_txt = ""
        for variable in variables:
            # Substitute the name and value of the environment variable from args
            env_template = f"epicsEnvSet {variable.name} {variable.value}"
            env_var_txt += UTILS.render(
                instance,
                env_template,
            )  # type: ignore
        return env_var_txt + "\n"

    def render_elements(
        self, ioc: IOC, method: Callable[[Entity], Union[str, None]]
    ) -> str:
        """
        Render elements of a given IOC instance based on calling the correct method
        """
        elements = ""
        for entity in ioc.entities:
            if entity.entity_enabled:
                element = method(entity)
                if element:
                    elements += element
        return elements

    def render_pre_ioc_init_elements(self, ioc: IOC) -> str:
        """
        Render all of the startup script entries for a given IOC instance
        """
        return self.render_elements(ioc, self.render_pre_ioc_init)

    def render_post_ioc_init_elements(self, ioc: IOC) -> str:
        """
        Render all of the post-iocInit elements for a given IOC instance
        """
        return self.render_elements(ioc, self.render_post_ioc_init)

    def render_environment_variable_elements(self, ioc: IOC) -> str:
        """
        Render all of the environment variable entries for a given IOC instance
        """
        return self.render_elements(ioc, self.render_environment_variables)



from __future__ import annotations

from typing import Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

id_to_entity: Dict[str, Entity] = {}


class Entity(BaseModel):
    name: str = Field(..., description="The name of this entity")
    value: str = Field(..., description="The value of this entity")
    ref: Optional[str] = Field(
        default=None, description="Reference another Entity name"
    )
    model_config = ConfigDict(extra="forbid")

    @model_validator(mode="after")  # type: ignore
    def add_ibek_attributes(cls, entity: Entity):
        id_to_entity[entity.name] = entity

        return entity

    @field_validator("ref", mode="after")
    def lookup_instance(cls, id):
        try:
            return id_to_entity[id]
        except KeyError:
            raise KeyError(f"object {id} not found in {list(id_to_entity)}")


class Entities(BaseModel):
    entities: List[Entity] = Field(..., description="The entities in this model")


model1 = Entities(
    **{
        "entities": [
            {"name": "one", "value": "OneValue"},
            {"name": "two", "value": "TwoValue", "ref": "one"},
        ]
    }
)

# demonstrate that entity two has a reference to entity one
assert model1.entities[1].ref.value == "OneValue"

# this should throw an error because entity one_again has illegal arguments
# BUT the error shown is:
#    KeyError: "object one_again not found in ['one', 'two']"
# which masks the underlying schema violation error that should look like:
#    Extra inputs are not permitted [type=extra_forbidden, input_value='bad argument',
model2 = Entities(
    **{
        "entities": [
            {"name": "one_again", "value": "OneValue", "illegal": "bad argument"},
            {"name": "two_again", "value": "TwoValue", "ref": "one_again"},
        ]
    }
)
